/****************************************************************************
**
** Copyright (C) 2020 The Qt Company Ltd.
** Contact: https://www.qt.io/licensing/
**
** This file is part of the QtQml module of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:LGPL$
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see https://www.qt.io/terms-conditions. For further
** information use the contact form at https://www.qt.io/contact-us.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 3 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL3 included in the
** packaging of this file. Please review the following information to
** ensure the GNU Lesser General Public License version 3 requirements
** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 2.0 or (at your option) the GNU General
** Public license version 3 or any later version approved by the KDE Free
** Qt Foundation. The licenses are as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
** included in the packaging of this file. Please review the following
** information to ensure the GNU General Public License requirements will
** be met: https://www.gnu.org/licenses/gpl-2.0.html and
** https://www.gnu.org/licenses/gpl-3.0.html.
**
** $QT_END_LICENSE$
**/
#ifndef QQMLDOM_FWD_P_H
#define QQMLDOM_FWD_P_H

//
//  W A R N I N G
//  -------------
//
// This file is not part of the Qt API.  It exists purely as an
// implementation detail.  This header file may change from version to
// version without notice, or even be removed.
//
// We mean it.
//

#include "qqmldom_global.h"

QT_BEGIN_NAMESPACE

namespace QQmlJS {
namespace Dom {

class ExternalItemInfoBase;
class ExternalItemPairBase;
class ExternalOwningItem;
class OwningItem;
class DomBase;
class DomItem;
class Source;
class Empty;
class QmlDirectory;
class JsFile;
class QmlFile;
class QmltypesFile;
class ModuleIndex;
class ModuleScope;
class Export;
class JsResource;
class QmltypesComponent;
class QmlComponent;
class EnumDecl;
class Import;
class Pragma;
class Id;
class QmlObject;
class ConstantData;
class ScriptExpression;
class Reference;
class Binding;
class PropertyDefinition;
class RequiredProperty;
class Version;
class MethodInfo;
class GenericObject;
class Map;
class List;
class Version;
class DomTop;
class DomEnvironment;
class DomUniverse;

class Subpath;
class ObserversTrie;

} // end namespace Dom
} // end namespace QQmlJS
QT_END_NAMESPACE
#endif // QQMLDOM_FWD_P_H
